import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision.models.resnet import resnet50, resnet18


class Model(nn.Module):
    def __init__(self, feature_dim=128):
        super(Model, self).__init__()

        self.f = []
        for name, module in resnet18().named_children():
            if name == 'conv1':
                module = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
            if not isinstance(module, nn.Linear) and not isinstance(module, nn.MaxPool2d):
                self.f.append(module)
        # encoder
        self.f = nn.Sequential(*self.f)
        # projection head
        self.g = nn.Sequential(nn.Linear(512, 512, bias=False), nn.BatchNorm1d(512),
                               nn.ReLU(inplace=True), nn.Linear(512, feature_dim, bias=True))

    def forward(self, x):
        x = self.f(x)
        feature = torch.flatten(x, start_dim=1)
        out = self.g(feature)
        return F.normalize(feature, dim=-1), F.normalize(out, dim=-1)



class DropoutModel(nn.Module):
    def __init__(self, feature_dim=128, dropout_rate=0.3):
        super(DropoutModel, self).__init__()

        self.f = []
        for name, module in resnet50().named_children():
            if name == 'conv1':
                module = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
            if not isinstance(module, nn.Linear) and not isinstance(module, nn.MaxPool2d):
                self.f.append(module)
                # Optionally add dropout layers within the feature extractor part
                if name in ['layer3', 'layer4']:  # Example positions for dropout
                    self.f.append(nn.Dropout(dropout_rate))

        # encoder
        self.f = nn.Sequential(*self.f)

        # projection head
        self.g = nn.Sequential(nn.Linear(2048, 512, bias=False), nn.BatchNorm1d(512),
                               nn.ReLU(inplace=True), nn.Dropout(dropout_rate),  # Dropout in projection head
                               nn.Linear(512, feature_dim, bias=True))

    def forward(self, x, apply_dropout=False):
        # If apply_dropout is True, enable dropout during inference
        if apply_dropout:
            self._enable_dropout()

        x = self.f(x)
        feature = torch.flatten(x, start_dim=1)
        out = self.g(feature)
        return F.normalize(feature, dim=-1), F.normalize(out, dim=-1)

    def _enable_dropout(self):
        # Function to enable dropout layers
        for module in self.modules():
            if isinstance(module, nn.Dropout):
                module.train()
